# internal
from .consts import EXTENSIONS

# dj
from django.core.exceptions import ImproperlyConfigured


def get_extension(output_format: str) -> str:
    try:
        return f".{EXTENSIONS[output_format]}"
    except KeyError:
        raise ImproperlyConfigured(f"Invalid format: {output_format}")



# standard
import json
import hashlib


def cache_key_generator(
    report_name: str, data: dict, dpi: int, output_format: str
) -> str:
    s = report_name + json.dumps(data) + str(dpi) + output_format
    s = s.encode("utf8")
    return hashlib.md5(s).hexdigest()



# typing
from typing import Any

# dj
from django.conf import settings
from django.utils.module_loading import import_string


class DJReportConf(object):
    """DJReport Conf"""

    @staticmethod
    def _get_settings(key: str, default: Any = None) -> Any:
        return getattr(settings, key, default)

    @property
    def installed(self) -> bool:
        return "djreport" in settings.INSTALLED_APPS

    @property
    def cache_key_generator(self) -> callable:
        dotted_path = self._get_settings(
            "DJREPORT_CACHE_KEY_GENERATOR", "djreport.utils.cache_key_generator"
        )
        return import_string(dotted_path)

    @property
    def filename_header_key(self) -> str:
        return self._get_settings("DJREPORT_FILENAME_HEADER_KEY", "X-FileName")

    @property
    def cache_timeout(self):
        return self._get_settings("DJREPORT_CACHE_TIMEOUT", 60 * 5)


djreport_conf = DJReportConf()



# standard
from typing import Any

# dj
from django.db import models
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

# internal
from .conf import djreport_conf
from .mixins import ReportDataSourceMixin
from .engine import Engine, ENGINE_CHOICES


class DataSource(models.Model):
    """Date Source"""

    name = models.CharField(max_length=150, unique=True)
    dotted_path = models.CharField(
        max_length=255, help_text="E.g. apps.accounting.data_sources.Invoice"
    )

    class Meta:
        abstract = not djreport_conf.installed

    @cached_property
    def instance(self) -> ReportDataSourceMixin:
        data_source_class = import_string(self.dotted_path)
        return data_source_class()

    def get_data(self, **kwargs) -> dict:
        return self.instance.get_data(**kwargs)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"DataSource(id={self.id}, name={self.name})"


class Report(models.Model):
    """Report Model"""

    active = models.BooleanField(default=True)
    _engine = models.CharField(
        verbose_name="Engine", max_length=50, choices=ENGINE_CHOICES
    )
    name = models.CharField(max_length=255, unique=True)
    file = models.FileField(upload_to="reports/")
    data_source = models.ForeignKey(
        "DataSource",
        related_name="reports",
        on_delete=models.CASCADE,
    )
    cache_required = models.BooleanField(default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = not djreport_conf.installed

    @cached_property
    def engine(self) -> Engine:
        return Engine(self._engine)

    def render(
        self, dpi: int, output_format: str, data: dict = None, **kwargs: Any
    ) -> bytes:
        # get data
        if data is None:
            data = self.data_source.get_data(**kwargs)
        # render
        return self.engine.render(self.file.path, data, dpi, output_format)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Report(id={self.id}, name={self.name})"



# dj
from django.core.cache import cache
from django.http import HttpResponse

# drf
from rest_framework import status
from rest_framework import renderers
from rest_framework.response import Response
from rest_framework.generics import GenericAPIView

# internal
from .models import Report
from .conf import djreport_conf
from .engine import get_extension
from .serializers import RenderSerializer


class RenderView(GenericAPIView):
    """Render View"""

    serializer_class = RenderSerializer
    queryset = Report.objects.filter(active=True)
    renderer_classes = [renderers.JSONRenderer]

    @staticmethod
    def render_report(
        report: Report, data: dict, dpi: int, output_format: str
    ) -> bytes:
        # check for report.cache_required
        if not report.cache_required:
            return report.render(dpi, output_format, data)
        # generate cache-key base on report.name, data, dpi and output_format
        key = djreport_conf.cache_key_generator(report.name, data, dpi, output_format)
        # get value by generated-key from cache registry
        value = cache.get(key)
        # if value does not exist:
        # 1) generate value by rendering report
        # 2) set value into cache registry for next time uses
        # 3) return value as result
        if not value:
            value = report.render(dpi, output_format, data)
            cache.set(key, value, djreport_conf.cache_timeout)

        return value

    @staticmethod
    def download_response(
        report_name: str, rendered_bytes: bytes, output_format: str
    ) -> HttpResponse:
        # create generic mime type file-response
        response = HttpResponse(rendered_bytes, content_type="application/octet-stream")
        # create file name base on given report_name and output_format
        filename = report_name + get_extension(output_format)
        # add filename to response headers as [filename_header_key] for local uses
        response[djreport_conf.filename_header_key] = filename
        # set Content-Disposition as attachment
        response["Content-Disposition"] = f'attachment; filename="{filename}"'
        return response

    def get(self, request, pk):  # noqa
        # get report
        report = self.get_object()
        # serialize and validate render parameters
        serializer = self.get_serializer(
            data=request.query_params, context={"report": report}
        )
        if serializer.is_valid():
            # extract dpi and output_format
            dpi = serializer.validated_data["dpi"]
            output_format = serializer.validated_data["output_format"]
            # get data from report.data_source
            data = report.data_source.get_data(**request.query_params)
            rendered_report = self.render_report(report, data, dpi, output_format)
            return self.download_response(report.name, rendered_report, output_format)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)



# dj
from django.contrib import admin

# internal
from .models import DataSource, Report


class DataSourceListFilter(admin.SimpleListFilter):
    """DataSource List Filter"""

    title = "data source"

    parameter_name = "data_source"

    def lookups(self, request, model_admin):
        return [(ds.id, ds.name) for ds in DataSource.objects.all()]

    def queryset(self, request, queryset):
        if value := self.value():
            queryset = queryset.filter(id=value)
        return queryset


@admin.register(DataSource)
class DataSourceAdmin(admin.ModelAdmin):
    """DataSource Admin"""

    list_display = ["id", "name", "dotted_path"]


@admin.register(Report)
class ReportAdmin(admin.ModelAdmin):
    """Report Admin"""

    list_display = ["id", "name", "data_source", "active"]
    list_filter = ["active", DataSourceListFilter]



# engines
STIMULSOFT = "stimulsoft"
FASTREPORT = "fastreport"

ENGINES = (
    STIMULSOFT,
    FASTREPORT,
)

ENGINE_CHOICES = ((STIMULSOFT, "Stimulsoft"), (FASTREPORT, "FastReport"))


# bins
STIMULSOFT_BIN = "DJREPORT_STIMULSOFT"
FASTREPORT_BIN = "DJREPORT_FASTREPORT"

BINS = {
    STIMULSOFT: STIMULSOFT_BIN,
    FASTREPORT: FASTREPORT_BIN,
}


# dpi
MIN_DPI = 96
MAX_DPI = 300


# formats
CSV = "csv"
DATA = "data"
DBF = "dbf"
DIF = "dif"
EXCEL = "excel"
HTML = "html"
HTML5 = "html5"
PDF = "pdf"
BMP = "bmp"
EMF = "emf"
GIF = "gif"
JPEG = "jpeg"
PCX = "pcx"
PNG = "png"
SVG = "svg"
SVGZ = "svgz"
TIFF = "tiff"
JSON = "json"
MHT = "mht"
ODS = "ods"
ODT = "odt"
POWER_POINT = "powerpoint"
RTF = "rtf"
SYLK = "sylk"
WORD = "word"
XML = "xml"
XPS = "xps"

FORMATS = (
    CSV,
    DATA,
    DBF,
    DIF,
    EXCEL,
    HTML,
    HTML5,
    PDF,
    BMP,
    EMF,
    GIF,
    JPEG,
    PCX,
    PNG,
    SVG,
    SVGZ,
    TIFF,
    JSON,
    MHT,
    ODS,
    ODT,
    POWER_POINT,
    RTF,
    SYLK,
    WORD,
    XML,
    XPS,
)


# extensions
EXTENSIONS = {
    CSV: "csv",
    DATA: "data",
    DBF: "dbf",
    DIF: "dif",
    EXCEL: "xls",
    HTML: "html",
    HTML5: "html",
    PDF: "pdf",
    BMP: "bmp",
    EMF: "emf",
    GIF: "gif",
    JPEG: "jpeg",
    PCX: "pcx",
    PNG: "png",
    SVG: "svg",
    SVGZ: "svgz",
    TIFF: "tiff",
    JSON: "json",
    MHT: "mht",
    ODS: "ods",
    ODT: "odt",
    POWER_POINT: "ppt",
    RTF: "rtf",
    SYLK: "slk",
    WORD: "doc",
    XML: "xml",
    XPS: "xps",
}



# Generated by Django 5.1.4 on 2024-12-17 16:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DataSource",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=150, unique=True)),
                (
                    "dotted_path",
                    models.CharField(
                        help_text="E.g. apps.accounting.data_sources.Invoice",
                        max_length=255,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Report",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "_engine",
                    models.CharField(
                        choices=[
                            ("stimulsoft", "Stimulsoft"),
                            ("fastreport", "FastReport"),
                        ],
                        max_length=50,
                        verbose_name="Engine",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("file", models.FileField(upload_to="reports/")),
                ("cache_required", models.BooleanField(default=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "data_source",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="reports",
                        to="djreport.datasource",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# drf
from rest_framework import serializers

# internal
from .engine import FORMATS, MIN_DPI, MAX_DPI


class RenderSerializer(serializers.Serializer):  # noqa
    """Render Serializer"""

    dpi = serializers.IntegerField(
        min_value=MIN_DPI, max_value=MAX_DPI, default=MIN_DPI
    )
    output_format = serializers.ChoiceField(choices=FORMATS)

    def validate_output_format(self, value):
        # get report instance from context
        report = self.context.get("report")
        # check that given report instance.engine support given output format
        if report:
            # validate given output format
            pass
        return value



# standard
import os
import json
import tempfile
import subprocess

# dj
from django.utils.functional import cached_property
from django.core.exceptions import ImproperlyConfigured

# internal
from .consts import BINS
from .utils import get_extension


class Engine(object):
    """Report Engine"""

    def __init__(self, engine: str) -> None:
        self._engine = engine

    @cached_property
    def bin(self) -> str:
        # get key
        try:
            key = BINS[self._engine]
        except KeyError:
            raise ImproperlyConfigured(f"Invalid engine: '{self._engine}'.")
        # get path
        path = os.getenv(key)
        if path is None:
            raise ImproperlyConfigured(f"'{key}' is not set.")
        # check for path
        if not os.path.exists(path):
            raise ImproperlyConfigured(f"'{path}' does not exist.")
        return path

    def render(
        self, report_path: str, data: dict, dpi: int, output_format: str
    ) -> bytes:
        # prepare data
        data = json.dumps(data, ensure_ascii=False)
        # get extension from given output_format
        extension = get_extension(output_format)
        # render
        with tempfile.NamedTemporaryFile(suffix=extension, delete=True) as f:
            subprocess.run(
                [self.bin, report_path, data, str(dpi), output_format, f.name],
                capture_output=True,
                text=True,
            )
            # rewind file
            f.seek(0)
            # read temp file content before delete
            content = f.read()
        # return content as result
        return content
